/*
 */

package com.erubioman.verilog.processor;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.erubioman.Loggable;

/*
 * ParsedVerilogSorter
 */
public class ParsedVerilogSorter implements Loggable {

	private class WireDetails {
		private Set<String> dependenciesSet;
		private Set<String> referencingSet;
		private ArrayList<String> assignmentList;
		private boolean beingProcessed;
		private boolean hasOperations;
		private boolean isWire;
		private boolean calculatingReferences;
		private boolean ignored;
		private Map<String, WireDetails> wireMap;

		public WireDetails(Map<String, WireDetails> wire_map, boolean is_wire) {
			dependenciesSet = new TreeSet<String>();
			referencingSet = new HashSet<String>();
			assignmentList = new ArrayList<String>();
			beingProcessed = false;
			hasOperations = false;
			isWire = is_wire;
			calculatingReferences = false;
			ignored = false;
			wireMap = wire_map;
		}

		public WireDetails(Map<String, WireDetails> wire_map, String equationRight) {
			this(wire_map, true);
			AddAssignment(equationRight);
		}

		public void AddAssignment(String equationRight) {
			String[] dependencies = equationRight.split("\\W");
			for (String dependency : dependencies) {
				dependenciesSet.add(dependency);
			}
			if ((dependenciesSet.size() > 1) || !equationRight.equals(dependencies[0])) {
				hasOperations = true;
			}
			assignmentList.add(equationRight);
		}

		public void SetProcessed(boolean new_value) {
			beingProcessed = new_value;
		}

		public boolean GetProcessed() {
			return beingProcessed;
		}

		public Set<String> GetDependencies() {
			return dependenciesSet;
		}

		public List<String> GetAssignments() {
			return assignmentList;
		}

		public void SetIsWire(boolean is_wire) {
			isWire = is_wire;
		}

		public boolean GetIsWire() {
			return isWire;
		}

		public String GetEqualWire() {
			return isWire && (!hasOperations) && (dependenciesSet.size() == 1) ? dependenciesSet.iterator().next()
					: null;
		}

		public void AddReference(String name) {
			referencingSet.add(name);
		}

		public int GetNumberOfReferences() {
			int result = 0;
			if (!isWire) {
				result = 1; // Es una sortida
			} else if (!calculatingReferences) {
				calculatingReferences = true;
				for (String referencing : referencingSet) {
					WireDetails referencingDetails = wireMap.get(referencing);
					result += (referencingDetails != null ? referencingDetails.GetNumberOfReferences() : 1);
				}
				calculatingReferences = false;
			}
			return result;
		}

		public boolean IsUsed() {
			return GetNumberOfReferences() > 0;
		}

		public void SetIgnored(boolean set_ignored) {
			ignored = set_ignored;
		}

		public boolean GetIgnored() {
			return ignored;
		}
	}

	private Map<String, WireDetails> wireMap;
	private Map<String, String> wireSubstitutionMap;
	private boolean doSortWires;
	private boolean simplifyWires;
	private boolean removeUnused;
	private String verilogOther;
	private Loggable parent;

	public ParsedVerilogSorter(Loggable _parent, boolean sort_wires, boolean simplify_wires, boolean remove_unused) {
		parent = _parent;
		wireMap = new TreeMap<String, WireDetails>(String.CASE_INSENSITIVE_ORDER);
		wireSubstitutionMap = new HashMap<String, String>();
		doSortWires = sort_wires;
		simplifyWires = simplify_wires;
		removeUnused = remove_unused;
		verilogOther = new String();
	}

	private void ProcessAssignment(String in, boolean avoidRepeated) {
		String[] equationParts = in.split("\\s*=\\s*");
		if (equationParts.length > 1) {
			String equationLeft = equationParts[0];
			String equationRight = equationParts[1];
			WireDetails details = wireMap.get(equationLeft);
			if (details == null) {
				wireMap.put(equationLeft, new WireDetails(wireMap, equationRight));
			} else if (!avoidRepeated) {
				details.AddAssignment(equationRight);
			}
		}
	}

	private void ProcessWireDeclaration(String in, boolean is_wire) {
		WireDetails details = wireMap.get(in);
		if (details == null) {
			wireMap.put(in, new WireDetails(wireMap, is_wire));
		} else {
			details.SetIsWire(is_wire);
		}
	}

	private static boolean StringContainsWord(String text, String word) {
		String source = "\\b" + word + "\\b";
		Pattern pattern = Pattern.compile(source);
		Matcher matcher = pattern.matcher(text);
		return matcher.find();
	}

	private void ProcessAssignments(String name, WireDetails details) {
		if ((details != null) && !details.GetProcessed()) {
			details.SetProcessed(true);
			if (StringContainsWord(verilogOther, name)) {
				details.AddReference("?");
			}
			for (String parent : details.GetDependencies()) {
				WireDetails parent_details = wireMap.get(parent);
				if (parent_details != null) {
					parent_details.AddReference(name);
					ProcessAssignments(parent, parent_details);
				}
			}
			String equal = (simplifyWires ? details.GetEqualWire() : null);
			if ((equal != null) && (wireMap.get(equal) != null)) {
				String parent = wireSubstitutionMap.get(equal);
				wireSubstitutionMap.put(name, parent != null ? parent : equal);
				details.SetIgnored(true);
			}
		}
	}

	private String OutputAssignments(String name, WireDetails details) {
		String output = new String();
		if ((details != null) && !details.GetProcessed()) {
			details.SetProcessed(true);
			for (String dependency : details.GetDependencies()) {
				output += OutputAssignments(dependency, wireMap.get(dependency));
			}
			if (!(simplifyWires && details.GetIgnored()) && !(removeUnused && !details.IsUsed())) {
				if (details.GetIsWire()) {
					output += ("wire " + name + ";\n");
				}
				String equationLeft = "assign " + name + " = ";
				for (String assignment : details.GetAssignments()) {
					output += (equationLeft + assignment + ";\n");
				}
			}
		}
		return output;
	}

	private String OutputAllAssignments() {
		// wireMap.forEach((k, v) -> ProcessAssignments(k, v));
		for (Map.Entry<String, WireDetails> entry : wireMap.entrySet()) {
			ProcessAssignments(entry.getKey(), entry.getValue());
		}
		// wireMap.forEach((k, v) -> v.SetProcessed(false));
		for (WireDetails details : wireMap.values()) {
			details.SetProcessed(false);
		}
		verilogOther = SubstituteWires(verilogOther);
		String output = new String();
		for (Map.Entry<String, WireDetails> entry : wireMap.entrySet()) {
			output += OutputAssignments(entry.getKey(), entry.getValue());
		}
		return SubstituteWires(output);
	}

	private String SubstituteWires(String in) {
		for (Map.Entry<String, String> entry : wireSubstitutionMap.entrySet()) {
			in = in.replaceAll("\\b" + entry.getKey() + "\\b", entry.getValue());
		}
		return in;
	}

	public String Translate(Readable in, String moduleName) {
		String output = new String();
		Scanner line_scanner = new Scanner(in);
		line_scanner.useDelimiter("\n");
		addLog("Canviant nom del mòdul: " + moduleName + "\n");

		if (doSortWires) {
			addLog("Ordenant dependències dels elements wire.\n");
			if (simplifyWires) {
				addLog("Eliminant elements wire redundants.\n");
			}
			if (removeUnused) {
				addLog("Eliminant elements wire no usats.\n");
			}
			String verilogDeclarations = new String();
			while (line_scanner.hasNext()) {
				String line = line_scanner.next();
				String[] sentences = line.split(";");
				if (sentences.length > 0) {
					String trimmed = sentences[0].trim();
					if (trimmed.startsWith("assign ")) {
						ProcessAssignment(trimmed.substring(7).trim(), false);
					} else if (trimmed.startsWith("wire ")) {
						ProcessWireDeclaration(trimmed.substring(5).trim(), true);
					} else if (trimmed.startsWith("input ")) {
						ProcessWireDeclaration(trimmed.substring(6).trim(), false);
						verilogDeclarations += (line + "\n");
					} else if (trimmed.startsWith("output ")) {
						ProcessWireDeclaration(trimmed.substring(7).trim(), false);
						verilogDeclarations += (line + "\n");
					} else if (trimmed.startsWith("reg ")) {
						ProcessWireDeclaration(trimmed.substring(4).trim(), false);
						verilogDeclarations += (line + "\n");
					} else if (trimmed.startsWith("module ")) {
						int pos = trimmed.indexOf('(');
						verilogDeclarations += "module " + moduleName + (pos < 0 ? "" : trimmed.substring(pos)) + ";\n";
						// verilogDeclarations += (line.replace('.', '_') +
						// "\n");
					} else {
						verilogOther += (line + "\n");
					}
				}
			}
			/* Avoid compiler errors caused by undefined symbols */
			ProcessAssignment("TRUE = 1", true);
			ProcessAssignment("FALSE = 0", true);
			output = verilogDeclarations + OutputAllAssignments() + verilogOther;
		} else {
			while (line_scanner.hasNext()) {
				String line = line_scanner.next();
				if (line.trim().startsWith("module ")) {
					int pos = line.indexOf('(');
					output += "module " + moduleName + (pos < 0 ? "" : line.substring(pos)) + "\n";
				} else {
					output += (line + "\n");
				}
			}
		}
		line_scanner.close();

		return output;
	}

	public boolean TranslateFiles(File inFile, File outFile) {
		boolean result = true;
		String fileName = inFile.getName();
		addLog("Convertint fitxer " + fileName + " a " + outFile.getName() + "...\n");
		try {
			String translated = Translate(new FileReader(inFile), "m_" + fileName.replaceAll("\\W", "_"));
			FileWriter writer = new FileWriter(outFile);
			writer.write(translated);
			writer.close();
		} catch (IOException ex) {
			result = false;
		}
		if (result) {
			addLog("Desat amb èxit.\n");
		} else {
			addLog("No es pot desar el fitxer.\n");
		}
		return result;

	}

	public void addLog(String text) {
		parent.addLog(text);
	}
}
