/*
 */

package com.erubioman.verilog.metrics;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringReader;
import java.nio.file.FileVisitOption;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import org.apache.commons.io.FilenameUtils;

import com.erubioman.Loggable;
import com.erubioman.verilog.processor.ParsedVerilogSorter;
import com.dbaneres.file.veriluoc.VerilogXMLWriter;

/*
 * VerilogAnalyzer
 */
public class VerilCircMetrics implements Loggable {

	private class Exercise implements Comparable<Exercise> {
		final String fullName;
		final String userName;
		final String semester;
		final String exerciseNum;
		final String numTry;
		final String date;
		final String time;

		@Override
		public int compareTo(Exercise o) {
			int result = exerciseNum.compareTo(o.exerciseNum);
			if (result == 0) {
				result = userName.compareTo(o.userName);
				if (result == 0) {
					result = semester.compareTo(o.semester);
					if (result == 0) {
						result = numTry.compareTo(o.numTry);
						if (result == 0) {
							result = fullName.compareTo(o.fullName);
						}
					}
				}
			}
			return result;
		}

		public Exercise(String _fullName, String _userName, String _semester, String _exerciseNum, String _numTry,
				String _date, String _time) {
			fullName = _fullName;
			userName = _userName;
			semester = _semester;
			exerciseNum = _exerciseNum;
			numTry = _numTry;
			date = _date;
			time = _time;
		}
	}

	private class ExerciseData {
		final String exerciseName;
		int teacherExerciseCount;
		int studentExerciseCount;

		ExerciseData(String name) {
			exerciseName = name;
			teacherExerciseCount = 0;
			studentExerciseCount = 0;
		}
	}

	private class FilesData {
		String circName;
		String verilogName;
	}

	private Set<Exercise> exercisesSet;
	private Map<String, Set<String>> mapNamesNumbers;
	private Map<String, ExerciseData> mapNumbersNames;
	private Map<String, FilesData> filesMap;
	private Map<String, Integer> cellDataMap = new TreeMap<String, Integer>();
	private final Loggable parent;
	private static final String verilogString = "verilog";
	private static final String tmpVerilogFileName = "temp_file.v";
	private static final String tmpResultsFileName = "temp_file.v.stats";
	private static final String yosysScript = "veriluoc_yosys_script.sh " + tmpVerilogFileName;
	private static final String yosysOptScript = "veriluoc_yosys_opt_script.sh " + tmpVerilogFileName;
	private static final String yosysResultsSearch[] = { "Number of wire bits:", "of processes:", "of cells:" };
	private static final String yosysResultsTitles[] = { "YosysWires", "YosysProcesses", "YosysCells" };
	private static final String yosysCellSearch[] = { "dff", "mux", "not", "and", "or" };
	private static final String yosysCellTitles[] = { "D_FlipFlops", "MuxGates", "InvGates", "AndGates", "OrGates" };
	private static final String abcResultsSearch[] = { "nd", "lev" };
	private static final String abcResultsTitles[] = { "AbcNodes", "AbcLevels" };

	VerilCircMetrics(Loggable _parent) {
		exercisesSet = new TreeSet<Exercise>();
		mapNamesNumbers = new TreeMap<String, Set<String>>();
		mapNumbersNames = new TreeMap<String, ExerciseData>();
		filesMap = new TreeMap<String, FilesData>();
		cellDataMap = new TreeMap<String, Integer>();
		parent = _parent;
	}

	private static boolean isACircFile(String fullName) {
		return FilenameUtils.getExtension(fullName).toLowerCase().equals("circ");
	}

	private static boolean isAVerilogFile(String fullName) {
		return verilogString.startsWith(FilenameUtils.getExtension(fullName).toLowerCase());
	}

	private boolean addFileToMap(Path file) {
		boolean added = false;
		String fileName = file.getFileName().toString();
		boolean isCirc = isACircFile(fileName);
		if (isCirc || isAVerilogFile(fileName)) {
			String withoutExt = FilenameUtils.removeExtension(fileName);
			FilesData fileData = filesMap.get(withoutExt);
			if (fileData == null) {
				fileData = new FilesData();
				filesMap.put(withoutExt, fileData);
			}
			if (isCirc) { /* Es un fitxer Circ */
				if (fileData.circName == null) {
					fileData.circName = file.toString();
					added = true;
				}
			} else { /* Es un fitxer Verilog */
				if (fileData.verilogName == null) {
					fileData.verilogName = file.toString();
					added = true;
				}
			}
		}
		if (!added) {
			addLog("Discarded file: " + file.toString() + "\n");
		}
		return added;
	}

	private int readExercisesInFilesMap() {
		int added = 0;
		for (String nomFitxer : filesMap.keySet()) {
			String parts[] = nomFitxer.split("_");
			int length = parts.length;
			if (length >= 6) { /* Es un exercici d'un alumne */
				String userName = parts[0];
				int nameLength = length - 5;
				for (int i = 1; i < nameLength; ++i) {
					userName += ("_") + parts[i];
				}
				Exercise exercise = new Exercise(nomFitxer, userName, parts[length - 5], parts[length - 4],
						parts[length - 3], parts[length - 2], parts[length - 1]);
				if (exercisesSet.add(exercise)) {
					++added;
					ExerciseData exerciseData = mapNumbersNames.get(exercise.exerciseNum);
					if (exerciseData == null) {
						exerciseData = new ExerciseData("-");
						mapNumbersNames.put(exercise.exerciseNum, exerciseData);
					}
					exerciseData.studentExerciseCount++;
				}
			} else { /* Es un exercici del professor */
				Set<String> exerciseNumbers = mapNamesNumbers.get(nomFitxer);
				if (exerciseNumbers == null) {
					exerciseNumbers = new TreeSet<String>();
					exerciseNumbers.add("-");
				}
				for (String number : exerciseNumbers) {
					Exercise exercise = new Exercise(nomFitxer, "0_Teacher", "", number, "", "", "");
					if (exercisesSet.add(exercise)) {
						++added;
						ExerciseData exerciseData = mapNumbersNames.get(exercise.exerciseNum);
						if (exerciseData == null) {
							exerciseData = new ExerciseData(nomFitxer);
							mapNumbersNames.put(exercise.exerciseNum, exerciseData);
						}
						exerciseData.teacherExerciseCount++;
					}
				}
			}
		}
		return added;
	}

	private int readCsvTable(File csvTable) {
		mapNumbersNames.clear();
		mapNamesNumbers.clear();
		int result = 0;
		if (csvTable != null) {
			addLog("Llegint taula d'exercicies del fitxer CSV: " + csvTable.getAbsolutePath() + "\n");
			try {
				Readable in = new FileReader(csvTable);
				Scanner line_scanner = new Scanner(in);
				line_scanner.useDelimiter("\n");
				while (line_scanner.hasNext()) {
					String line = line_scanner.next();
					String[] fields = line.split(";");
					if (fields.length >= 2) {
						String number = fields[0];
						String name = fields[1];
						Set<String> numbers = mapNamesNumbers.get(name);
						if (numbers == null) {
							numbers = new TreeSet<String>();
							mapNamesNumbers.put(name, numbers);
						}
						numbers.add(number);
						mapNumbersNames.put(number, new ExerciseData(name));
					}
				}
				line_scanner.close();
			} catch (IOException ex) {
				addLog(ex.getMessage());
			}
			result = mapNamesNumbers.size();
			if (result > 0) {
				addLog(result + " línies llegides.\n");
			} else {
				addLog("ERROR: " + csvTable.getAbsolutePath() + " no és una taula d'exercicis vàlida.\n");
			}
		}
		return result;
	}

	private int readExercisesInDirectory(File directory) {
		class Visitor extends SimpleFileVisitor<Path> {
			private VerilCircMetrics parent;
			private int fileCount;

			Visitor(VerilCircMetrics _parent) {
				parent = _parent;
				fileCount = 0;
			}

			@Override
			public FileVisitResult visitFile(Path file, BasicFileAttributes attr) {
				if (attr.isRegularFile()) {
					if (parent.addFileToMap(file)) {
						++fileCount;
					}
				}
				return FileVisitResult.CONTINUE;
			}

			int getFileCount() {
				return fileCount;
			}
		}
		Visitor visitor = new Visitor(this);
		addLog("Reading exercises from directory " + directory.getAbsolutePath() + "\n");
		if (directory.isDirectory()) {
			try {
				Files.walkFileTree(directory.toPath(), EnumSet.noneOf(FileVisitOption.class), Integer.MAX_VALUE,
						visitor);
			} catch (IOException ex) {
				addLog(ex.getMessage());
			}
		}
		final int fileCount = visitor.getFileCount();
		if (fileCount > 0) {
			addLog(fileCount + " exercises found.\n");
		} else {
			addLog("ERROR: no exercises found.\n");
		}
		return fileCount;
	}

	private ArrayList<String> dumpYosysCellStats(String[] yosysFields, String[] cellFields, String cellStats) {
		Scanner in = new Scanner(new StringReader(cellStats));
		in.useDelimiter("\\s+=*\\s*");
		ArrayList<String> result = new ArrayList<String>();
		for (String field : yosysFields) {
			if ((in.findWithinHorizon(field, 0) != null) && in.hasNext()) {
				String foundValue = in.next();
				result.add(foundValue);
			}
		}
		int expectedSize = yosysFields.length;
		for (int i = result.size(); i < expectedSize; ++i) {
			result.add("");
		}
		Map<String, Integer> cellData = new TreeMap<String, Integer>();
		for (String cellType : cellFields) {
			cellData.put(cellType, 0);
		}
		while ((in.findWithinHorizon("\\$", 0) != null) && in.hasNext()) {
			String name = in.next().toLowerCase();
			Integer cellTypeCount = cellDataMap.get(name);
			cellDataMap.put(name, cellTypeCount == null ? 1 : cellTypeCount + 1);
			if (in.hasNextInt()) {
				for (String cellType : cellFields) {
					if (name.contains(cellType)) {
						int foundValue = in.nextInt();
						cellData.put(cellType, cellData.get(cellType) + foundValue);
						break;
					}
				}
			}
		}
		in.close();
		for (String cellType : cellFields) {
			result.add(cellData.get(cellType).toString());
		}
		return result;
	}

	private ArrayList<String> executeYosysScript(String script, String[] yosysFields, String[] cellFields,
			String[] abcFields) {
		ArrayList<String> results = new ArrayList<String>();
		try {
			Process process = Runtime.getRuntime().exec(script);
			process.waitFor();
			Scanner in = new Scanner(new FileReader(tmpResultsFileName));
			/*
			 * Cerca els resultats de la primera estadistica (abans d'executar
			 * proc)
			 */
			in.useDelimiter("Printing statistics.");
			if ((in.findWithinHorizon("Printing statistics.", 0) != null) && in.hasNext()) {
				ArrayList<String> cellResult = dumpYosysCellStats(yosysFields, cellFields, in.next());
				/*
				 * Cerca els resultats de la segona estadistica (despres
				 * d'executar proc i, opcionalment, opt
				 */
				if (in.hasNext()) {
					in.next();
					if (in.hasNext()) {
						cellResult = dumpYosysCellStats(yosysFields, cellFields, in.next());
					}
				}
				results.addAll(cellResult);
				/* Cerca els resultats de l'execucio d'ABC */
				in.useDelimiter("\\s+=*\\s*");
				if (in.findWithinHorizon("Don't call ABC as there is nothing to map.", 0) != null) {
					for (int i = abcFields.length; i > 0; --i) {
						results.add("0");
					}
				} else if (in.findWithinHorizon("ABC: netlist", 0) != null) {
					for (String field : abcFields) {
						if ((in.findInLine(field) != null) && in.hasNext()) {
							results.add(in.next());
						}
					}
				}
			}
			in.close();
		} catch (IOException e) {
			addLog("ERROR: error d'E/S executant " + script + ":\n" + e.getMessage() + "\n");
		} catch (InterruptedException e) {
			addLog("ERROR: " + script + " interromput:\n" + e.getMessage() + "\n");
		}
		int expectedSize = yosysFields.length + cellFields.length + abcFields.length;
		for (int i = results.size(); i < expectedSize; ++i) {
			results.add("");
		}
		return results;
	}

	private ArrayList<String> getResultsFromTempFile(String[] yosysFields, String[] cellFields, String[] abcFields) {
		ArrayList<String> results = executeYosysScript(yosysScript, yosysFields, cellFields, abcFields);
		results.addAll(executeYosysScript(yosysOptScript, yosysFields, cellFields, abcFields));
		try {
			Files.deleteIfExists(new File(tmpVerilogFileName).toPath());
			Files.deleteIfExists(new File(tmpResultsFileName).toPath());
		} catch (IOException e) {
			addLog("ERROR: no es pot esborrar fitxer:\n" + e.getMessage() + "\n");
		}
		int size = results.size();
		addLog(size + " mètriques trobades.\n");
		return results;
	}

	private ArrayList<String> getResultsFromVerilogFile(String fullPath, String[] yosysFields, String[] cellFields,
			String[] abcFields) {
		addLog("Processant fitxer " + fullPath + "\n");
		new ParsedVerilogSorter(this, true, true, false).TranslateFiles(new File(fullPath),
				new File(tmpVerilogFileName));
		return getResultsFromTempFile(yosysFields, cellFields, abcFields);
	}

	private ArrayList<String> getResultsFromCircFile(String circFullPath, String[] yosysFields, String[] cellFields,
			String[] abcFields, String exerciseName) {
		boolean converted = false;
		addLog("Intentant processar fitxer CIRC.\n");
		addLog("Creant Verilog a partir de fitxer CIRC: " + circFullPath + "\n");
		String moduleName = "m_" + exerciseName.replaceAll("\\W", "_");
		String verilogResult = new String();
		try {
			/* Filetype should be 1 by default, 0 for older versions */
			verilogResult = new VerilogXMLWriter().initializeV(new File(circFullPath), 1);
			if (verilogResult.contains("error_3_[")) {
				addLog("Format VerilUOC causa error_3. Provant format Logisim...\n");
				verilogResult = new VerilogXMLWriter().initializeV(new File(circFullPath), 0);
			}
		} catch (Exception ex) {
			verilogResult = "Error de parsejat: fitxer CIRC incorrecte?\n" + ex.getMessage() + "\n";
			addLog(verilogResult);
		}
		if (verilogResult.startsWith("module") && !verilogResult.contains("error_3_[")) {
			try {
				FileWriter writer = new FileWriter(tmpVerilogFileName);
				writer.write(new ParsedVerilogSorter(this, true, true, false).Translate(new StringReader(verilogResult),
						moduleName));
				writer.close();
				converted = true;
			} catch (IOException ex) {
				addLog("No es pot escriure arxiu Verilog temporal!\n");
			}
		} else {
			addLog("Traducció de CIRC fallida: " + verilogResult + "\n");
		}
		return converted ? getResultsFromTempFile(yosysFields, cellFields, abcFields) : new ArrayList<String>();
	}

	private static ArrayList<String> getResultTitles() {
		ArrayList<String> result = new ArrayList<String>();
		for (String writeResultTitle : yosysResultsTitles) {
			result.add(writeResultTitle);
		}
		for (String writeResultTitle : yosysCellTitles) {
			result.add(writeResultTitle);
		}
		for (String writeResultTitle : abcResultsTitles) {
			result.add(writeResultTitle);
		}
		for (String writeResultTitle : yosysResultsTitles) {
			result.add("Opt" + writeResultTitle);
		}
		for (String writeResultTitle : yosysCellTitles) {
			result.add("Opt" + writeResultTitle);
		}
		for (String writeResultTitle : abcResultsTitles) {
			result.add("Opt" + writeResultTitle);
		}
		return result;
	}

	private int outputTable(File outFile, String namesRegEx) {
		addLog("Escrivint resultats al fitxer CSV: " + outFile.getAbsolutePath() + "\n");
		int exercisesNumber = 0;
		try {
			FileWriter out = new FileWriter(outFile);
			String resultTitles = new String("");
			for (String writeResultTitle : getResultTitles()) {
				resultTitles += ";" + writeResultTitle;
			}
			out.write("ExerciseNumber;UserName;Semester;Try;Date;Time;ExerciseName;CircPath;VerilogPath"
					+ ";TeacherSolutions;StudentSolutions" + resultTitles + ";ResultType\n");
			for (Exercise exercise : exercisesSet) {
				String num = exercise.exerciseNum;
				ExerciseData exerciseData = mapNumbersNames.get(num);
				if ((exerciseData != null) && ((namesRegEx == null) || ((num != null) && num.matches(namesRegEx)))) {
					FilesData filesData = filesMap.get(exercise.fullName);
					String circPath = filesData != null ? filesData.circName : null;
					String verilogPath = filesData != null ? filesData.verilogName : null;
					String line = num + ";" + exercise.userName + ";" + exercise.semester + ";" + exercise.numTry + ";"
							+ exercise.date + ";" + exercise.time + ";" + exerciseData.exerciseName + ";" + circPath
							+ ";" + verilogPath + ";" + exerciseData.teacherExerciseCount + ";"
							+ exerciseData.studentExerciseCount;
					ArrayList<String> results = new ArrayList<String>();
					String resultType = "None";
					if (circPath != null) {
						results = getResultsFromCircFile(circPath, yosysResultsSearch, yosysCellSearch,
								abcResultsSearch, exercise.fullName);
						if (results.size() > 0) {
							resultType = "Circ";
						}
					}
					if ((results.size() == 0) && (verilogPath != null)) {
						results = getResultsFromVerilogFile(verilogPath, yosysResultsSearch, yosysCellSearch,
								abcResultsSearch);
						if (results.size() > 0) {
							resultType = "Verilog";
						}
					}
					for (String result : results) {
						line += ";" + result;
					}
					int expectedLength = (yosysResultsSearch.length + yosysCellSearch.length + abcResultsSearch.length)
							* 2;
					for (int i = results.size(); i < expectedLength; ++i) {
						line += ";";
					}
					out.write(line + ";" + resultType + "\n");
					++exercisesNumber;
				}
			}
			out.close();
		} catch (IOException ex) {
			addLog(ex.getMessage());
		}
		addLog(exercisesNumber + " exercicis escrits.\n");
		addLog("Tipus de cel.les trobats:\n");
		for (Map.Entry<String, Integer> cellData : cellDataMap.entrySet()) {
			addLog(cellData.getKey() + ": " + cellData.getValue().toString() + "\n");
		}
		return exercisesNumber;
	}

	private boolean readOneExercise(File file) {
		ArrayList<String> results = new ArrayList<String>();
		String name = file.getName();
		if (isACircFile(name)) {
			results = getResultsFromCircFile(file.toString(), yosysResultsSearch, yosysCellSearch, abcResultsSearch,
					FilenameUtils.removeExtension(name));
		} else if (isAVerilogFile(name)) {
			results = getResultsFromVerilogFile(file.toString(), yosysResultsSearch, yosysCellSearch, abcResultsSearch);
		}
		ArrayList<String> titles = getResultTitles();
		int length = Math.min(results.size(), titles.size());
		if (length > 0) {
			addLog("\n----- Resultats:\n");
			for (int i = 0; i < length; ++i) {
				addLog(titles.get(i) + " = " + results.get(i) + "\n");
			}
			addLog("----- Fi dels resultats.\n\n");
		}
		return (length > 0);
	}

	public int analyzeFiles(File csvTable, String[] exercisesDirectories, File outFile, String namesRegEx) {
		int fileCount = 0;
		readCsvTable(csvTable);
		for (String directory : exercisesDirectories) {
			File file = new File(directory);
			if (file.isDirectory()) {
				readExercisesInDirectory(file);
			} else if (file.isFile()) {
				if (readOneExercise(file)) {
					++fileCount;
				}
			}
		}
		readExercisesInFilesMap();
		return (exercisesSet.size() > 0 ? outputTable(outFile, namesRegEx) : fileCount);
	}

	public void addLog(String text) {
		parent.addLog(text);
	}
}
